package net.minestom.server.inventory.click.integration;

import net.minestom.server.api.Env;
import net.minestom.server.api.EnvTest;
import net.minestom.server.coordinate.Pos;
import net.minestom.server.entity.Player;
import net.minestom.server.event.inventory.InventoryPreClickEvent;
import net.minestom.server.inventory.Inventory;
import net.minestom.server.inventory.InventoryType;
import net.minestom.server.inventory.click.ClickType;
import net.minestom.server.item.ItemStack;
import net.minestom.server.item.Material;
import net.minestom.server.network.packet.client.play.ClientClickWindowPacket;
import net.minestom.server.utils.inventory.PlayerInventoryUtils;
import org.junit.jupiter.api.Test;

import java.util.ArrayList;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

@EnvTest
public class DragRightClickIntegrationTest {

    @Test
    public void rightDragSelf(Env env) {
        var instance = env.createFlatInstance();
        var player = env.createPlayer(instance, new Pos(0, 40, 0));
        var inventory = player.getInventory();
        var listener = env.listen(InventoryPreClickEvent.class);
        // Drag to air
        {
            inventory.setCursorItem(ItemStack.of(Material.STONE, 64));
            List<InventoryPreClickEvent> events = new ArrayList<>();
            listener.followup(events::add);
            dragRight(player, List.of(1, 2));
            // start->slot->slot->end
            assertEquals(4, events.size());
            {
                // Start
                var event = events.get(0);
                assertNull(event.getInventory());
                assertEquals(ClickType.START_RIGHT_DRAGGING, event.getClickType());
                assertEquals(-999, event.getSlot());
                assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                assertEquals(ItemStack.AIR, event.getClickedItem());
            }
            {
                // Slot 1
                var event = events.get(1);
                assertNull(event.getInventory());
                assertEquals(ClickType.RIGHT_DRAGGING, event.getClickType());
                assertEquals(1, event.getSlot());
                assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                assertEquals(ItemStack.AIR, event.getClickedItem());
            }
            {
                // Slot 2
                var event = events.get(2);
                assertNull(event.getInventory());
                assertEquals(ClickType.RIGHT_DRAGGING, event.getClickType());
                assertEquals(2, event.getSlot());
                assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                assertEquals(ItemStack.AIR, event.getClickedItem());
            }
            {
                // End
                var event = events.get(3);
                assertNull(event.getInventory());
                assertEquals(ClickType.END_RIGHT_DRAGGING, event.getClickType());
                assertEquals(-999, event.getSlot());
                assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                assertEquals(ItemStack.AIR, event.getClickedItem());
            }

            assertEquals(ItemStack.of(Material.STONE, 62), inventory.getCursorItem());
            assertEquals(ItemStack.of(Material.STONE, 1), inventory.getItemStack(1));
            assertEquals(ItemStack.of(Material.STONE, 1), inventory.getItemStack(2));
        }
        // Cancel start
        {
            inventory.clear();
            inventory.setCursorItem(ItemStack.of(Material.STONE, 64));
            listener.followup(event -> {
                if (event.getClickType() == ClickType.START_RIGHT_DRAGGING) {
                    assertEquals(-999, event.getSlot());
                    assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                    assertEquals(ItemStack.AIR, event.getClickedItem());
                    event.setCancelled(true);
                } else {
                    fail("Start drag has been cancelled, following events should not be called");
                }
            });
            dragRight(player, List.of(1, 2));
            assertEquals(ItemStack.of(Material.STONE, 64), inventory.getCursorItem());
            assertEquals(ItemStack.AIR, inventory.getItemStack(1));
            assertEquals(ItemStack.AIR, inventory.getItemStack(2));
        }
        // Cancel end
        {
            inventory.clear();
            inventory.setCursorItem(ItemStack.of(Material.STONE, 64));
            listener.followup(event -> {
                if (event.getClickType() == ClickType.END_RIGHT_DRAGGING) {
                    assertEquals(-999, event.getSlot());
                    assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                    assertEquals(ItemStack.AIR, event.getClickedItem());
                    event.setCancelled(true);
                }
            });
            dragRight(player, List.of(1, 2));
            assertEquals(ItemStack.of(Material.STONE, 64), inventory.getCursorItem());
            assertEquals(ItemStack.AIR, inventory.getItemStack(1));
            assertEquals(ItemStack.AIR, inventory.getItemStack(2));
        }
        // Cancel step
        {
            inventory.clear();
            inventory.setCursorItem(ItemStack.of(Material.STONE, 64));
            listener.followup(event -> {
                if (event.getClickType() == ClickType.RIGHT_DRAGGING && event.getSlot() == 2) {
                    assertEquals(2, event.getSlot());
                    assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                    assertEquals(ItemStack.AIR, event.getClickedItem());
                    event.setCancelled(true);
                }
            });
            dragRight(player, List.of(1, 2, 3));
            assertEquals(ItemStack.of(Material.STONE, 62), inventory.getCursorItem());
            assertEquals(ItemStack.of(Material.STONE, 1), inventory.getItemStack(1));
            assertEquals(ItemStack.AIR, inventory.getItemStack(2));
            assertEquals(ItemStack.of(Material.STONE, 1), inventory.getItemStack(3));
        }
    }

    @Test
    public void rightDragExternal(Env env) {
        var instance = env.createFlatInstance();
        var player = env.createPlayer(instance, new Pos(0, 40, 0));
        var inventory = new Inventory(InventoryType.HOPPER, "test");
        player.openInventory(inventory);
        var listener = env.listen(InventoryPreClickEvent.class);
        // Drag to air
        {
            inventory.setCursorItem(player, ItemStack.of(Material.STONE, 64));
            List<InventoryPreClickEvent> events = new ArrayList<>();
            listener.followup(events::add);
            dragRightOpenInventory(player, List.of(1, 2));
            // start->slot->slot->end
            assertEquals(4, events.size());
            {
                // Start
                var event = events.get(0);
                assertNull(event.getInventory());
                assertEquals(ClickType.START_RIGHT_DRAGGING, event.getClickType());
                assertEquals(-999, event.getSlot());
                assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                assertEquals(ItemStack.AIR, event.getClickedItem());
            }
            {
                // Slot 1
                var event = events.get(1);
                assertEquals(inventory, event.getInventory());
                assertEquals(ClickType.RIGHT_DRAGGING, event.getClickType());
                assertEquals(1, event.getSlot());
                assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                assertEquals(ItemStack.AIR, event.getClickedItem());
            }
            {
                // Slot 2
                var event = events.get(2);
                assertEquals(inventory, event.getInventory());
                assertEquals(ClickType.RIGHT_DRAGGING, event.getClickType());
                assertEquals(2, event.getSlot());
                assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                assertEquals(ItemStack.AIR, event.getClickedItem());
            }
            {
                // End
                var event = events.get(3);
                assertNull(event.getInventory());
                assertEquals(ClickType.END_RIGHT_DRAGGING, event.getClickType());
                assertEquals(-999, event.getSlot());
                assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                assertEquals(ItemStack.AIR, event.getClickedItem());
            }

            assertEquals(ItemStack.of(Material.STONE, 62), inventory.getCursorItem(player));
            assertEquals(ItemStack.of(Material.STONE, 1), inventory.getItemStack(1));
            assertEquals(ItemStack.of(Material.STONE, 1), inventory.getItemStack(2));
        }
        // Cancel start
        {
            inventory.clear();
            inventory.setCursorItem(player, ItemStack.of(Material.STONE, 64));
            listener.followup(event -> {
                if (event.getClickType() == ClickType.START_RIGHT_DRAGGING) {
                    assertEquals(-999, event.getSlot());
                    assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                    assertEquals(ItemStack.AIR, event.getClickedItem());
                    event.setCancelled(true);
                } else {
                    fail("Start drag has been cancelled, following events should not be called");
                }
            });
            dragRightOpenInventory(player, List.of(1, 2));
            assertEquals(ItemStack.of(Material.STONE, 64), inventory.getCursorItem(player));
            assertEquals(ItemStack.AIR, inventory.getItemStack(1));
            assertEquals(ItemStack.AIR, inventory.getItemStack(2));
        }
        // Cancel end
        {
            inventory.clear();
            inventory.setCursorItem(player, ItemStack.of(Material.STONE, 64));
            listener.followup(event -> {
                if (event.getClickType() == ClickType.END_RIGHT_DRAGGING) {
                    assertEquals(-999, event.getSlot());
                    assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                    assertEquals(ItemStack.AIR, event.getClickedItem());
                    event.setCancelled(true);
                }
            });
            dragRightOpenInventory(player, List.of(1, 2));
            assertEquals(ItemStack.of(Material.STONE, 64), inventory.getCursorItem(player));
            assertEquals(ItemStack.AIR, inventory.getItemStack(1));
            assertEquals(ItemStack.AIR, inventory.getItemStack(2));
        }
        // Cancel step
        {
            inventory.clear();
            inventory.setCursorItem(player, ItemStack.of(Material.STONE, 64));
            listener.followup(event -> {
                if (event.getClickType() == ClickType.RIGHT_DRAGGING && event.getSlot() == 2) {
                    assertEquals(2, event.getSlot());
                    assertEquals(ItemStack.of(Material.STONE, 64), event.getCursorItem());
                    assertEquals(ItemStack.AIR, event.getClickedItem());
                    event.setCancelled(true);
                }
            });
            dragRightOpenInventory(player, List.of(1, 2, 3));
            assertEquals(ItemStack.of(Material.STONE, 62), inventory.getCursorItem(player));
            assertEquals(ItemStack.of(Material.STONE, 1), inventory.getItemStack(1));
            assertEquals(ItemStack.AIR, inventory.getItemStack(2));
            assertEquals(ItemStack.of(Material.STONE, 1), inventory.getItemStack(3));
        }
    }

    private void dragRightOpenInventory(Player player, List<Integer> slots) {
        _dragRight(player.getOpenInventory(), true, player, slots);
    }

    private void dragRight(Player player, List<Integer> slots) {
        _dragRight(player.getOpenInventory(), false, player, slots);
    }

    private void _dragRight(Inventory openInventory, boolean clickOpenInventory, Player player, List<Integer> slots) {
        final byte windowId = openInventory != null ? openInventory.getWindowId() : 0;

        // Start right drag
        player.addPacketToQueue(new ClientClickWindowPacket(windowId, 0, (short) -999, (byte) 4,
                ClientClickWindowPacket.ClickType.QUICK_CRAFT, List.of(), ItemStack.AIR));
        // Add all slots
        for (int slot : slots) {
            if (clickOpenInventory) {
                assert openInventory != null;
                // Do not touch slot
            } else {
                int offset = openInventory != null ? openInventory.getInnerSize() : 0;
                slot = PlayerInventoryUtils.convertToPacketSlot(slot);
                if (openInventory != null) {
                    slot = slot - 9 + offset;
                }
            }
            player.addPacketToQueue(new ClientClickWindowPacket(windowId, 0, (short) slot, (byte) 5,
                    ClientClickWindowPacket.ClickType.QUICK_CRAFT, List.of(), ItemStack.AIR));
        }
        // End right drag
        player.addPacketToQueue(new ClientClickWindowPacket(windowId, 0, (short) -999, (byte) 6,
                ClientClickWindowPacket.ClickType.QUICK_CRAFT, List.of(), ItemStack.AIR));

        player.interpretPacketQueue();
    }
}
